/* ldapmodify.c - generic program to modify or add entries using LDAP */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#ifndef VMS
#include <unistd.h>
#endif /* VMS */
#include <lber.h>
#include <ldap.h>

#include "ldapconfig.h"

static char	*binddn = LDAPMODIFY_BINDDN;
static char	*base = LDAPMODIFY_BASE;
static char	*passwd = NULL;
static char	*ldaphost = LDAPHOST;
static int	ldapport = LDAP_PORT;
static int	new, replace, not, verbose, contoper, valsfromfiles;
static LDAP	*ld;

#ifdef LDAP_DEBUG
extern int ldap_debug, lber_debug;
#endif /* LDAP_DEBUG */

#define safe_realloc( ptr, size )	( ptr == NULL ? malloc( size ) : \
					 realloc( ptr, size ))


main( argc, argv )
    int		argc;
    char	**argv;
{
    char		*prog, *infile, *p, *bp, *eolp, *entrydn, buf[ 4096 ];
    FILE		*fp;
    int			rc, i, kerberos, linenum, bytes;
    LDAPMod		**pmods;
    char		*usage = "usage: %s [-abcknrv] [-d debug-level] [-h ldaphost] [-p ldapport] [-D binddn] [-w passwd] [ -f file | < entryfile ]\n";

    extern char	*optarg;
    extern int	optind;

    if (( prog = strrchr( argv[ 0 ], '/' )) == NULL ) {
	prog = argv[ 0 ];
    } else {
	++prog;
    }
    new = ( strcmp( prog, "ldapadd" ) == 0 );

    infile = NULL;
    kerberos = not = verbose = valsfromfiles = 0;

    while (( i = getopt( argc, argv, "abcknrtvh:p:D:w:d:f:" )) != EOF ) {
	switch( i ) {
	case 'a':	/* add */
	    new = 1;
	    break;
	case 'b':	/* read values from files (for binary attributes) */
	    valsfromfiles = 1;
	    break;
	case 'c':	/* continuous operation */
	    contoper = 1;
	    break;
	case 'r':	/* default is to replace rather than add values */
	    replace = 1;
	    break;
	case 'k':	/* kerberos bind */
	    kerberos = 1;
	    break;
	case 'h':	/* ldap host */
	    ldaphost = strdup( optarg );
	    break;
	case 'D':	/* bind DN */
	    binddn = strdup( optarg );
	    break;
	case 'w':	/* password */
	    passwd = strdup( optarg );
	    break;
	case 'd':
#ifdef LDAP_DEBUG
	    ldap_debug = lber_debug = atoi( optarg );	/* */
#else /* LDAP_DEBUG */
	    fprintf( stderr, "compile with -DLDAP_DEBUG for debugging\n" );
#endif /* LDAP_DEBUG */
	    break;
	case 'f':	/* read from file */
	    infile = strdup( optarg );
	    break;
	case 'p':
	    ldapport = atoi( optarg );
	    break;
	case 'n':	/* print adds, don't actually do them */
	    ++not;
	    break;
	case 'v':	/* verbose mode */
	    verbose++;
	    break;
	default:
	    fprintf( stderr, usage, prog );
	    exit( 1 );
	}
    }

    if ( argc - optind != 0 ) {
	fprintf( stderr, usage, prog );
	exit( 1 );
    }

    if ( infile != NULL ) {
	if (( fp = fopen( infile, "r" )) == NULL ) {
	    perror( infile );
	    exit( 1 );
	}
    } else {
	fp = stdin;
    }

    if (( ld = ldap_open( ldaphost, ldapport )) == NULL ) {
	perror( "ldap_open" );
	exit( 1 );
    }

    ld->ld_deref = LDAP_DEREF_NEVER;	/* this seems prudent */

    if ( ldap_bind_s( ld, binddn, passwd,
	    kerberos ? LDAP_AUTH_KRBV4 : LDAP_AUTH_SIMPLE ) != LDAP_SUCCESS ) {
	ldap_perror( ld, "ldap_bind" );
	exit( 1 );
    }

    rc = 0;
    linenum = 0;
    pmods = NULL;
    bp = buf;
    bytes = sizeof(buf);

    while (( rc == 0 || contoper ) && fgets( bp, bytes, fp ) != NULL ) {
	++linenum;

	if (( eolp = strrchr( bp, '\n' )) == NULL ) {
	    fprintf( stderr, "Line %d too long\n", linenum );
	    exit( 1 );
	}

	*eolp = '\0';
	--eolp;

	if ( *bp == '\0' ) {	/* empty line -- end one entry's modify ops */
	    if ( pmods != NULL ) {
		rc = domodify( ld, entrydn, pmods );
		freepmods( pmods );
		pmods = NULL;
	    }
	    bp = buf;
	    bytes = sizeof(buf);
	    linenum = 0;
	} else if ( *eolp == '\\' ) {	/* lines ending in '\' are continued */
	    *eolp++ = '\n';
	    bytes -= (eolp - bp);
	    bp = eolp;
	    continue;
	} else if ( linenum == 1 ) {	/* first line contains DN */
	    if (( entrydn = strdup( buf )) == NULL ) {
		perror( "strdup" );
		exit( 1 );
	    }
	    bp = buf;
	    bytes = sizeof(buf);
	} else {
	    addmodifyop( &pmods, buf, linenum );
	    bp = buf;
	    bytes = sizeof(buf);
	}
    }

    if ((rc == 0 || contoper) && pmods != NULL) {
	rc = domodify( ld, entrydn, pmods );
	freepmods( pmods );
	pmods = NULL;
    }

    ldap_unbind( ld );

    exit( rc );
}


addmodifyop( pmodsp, buf, linenum )
    LDAPMod	***pmodsp;
    char	*buf;
    int		linenum;
{
    char		*p, *q, *attr, *value;
    LDAPMod		**pmods;
    int			i, j, modop;
    struct berval	*bvp;
		    
    pmods = *pmodsp;

    if (( p = strchr( buf, '=' )) == NULL ) {
	value = NULL;
	p = buf + strlen( buf );
    } else {
	*p++ = '\0';
	value = p;
    }

    for ( attr = buf; *attr != '\0' && isspace( *attr ); ++attr ) {
	;	/* skip attribute leading white space */
    }

    for ( q = p - 1; q > attr && isspace( *q ); --q ) {
	*q = '\0';	/* remove attribute trailing white space */
    }

    if ( value != NULL ) {
	while ( isspace( *value )) {
	    ++value;		/* skip value leading white space */
	}
	for ( q = value + strlen( value ) - 1; q > value && isspace( *q );
		--q ) {
	    *q = '\0';	/* remove value trailing white space */
	}
    } else if ( new ) {
	fprintf( stderr, "missing value on line %d (attr is %s):\n",
		linenum, attr );
	exit( 1 );
    }

    switch ( *attr ) {
    case '-':
	modop = LDAP_MOD_DELETE;
	++attr;
	break;
    case '+':
	modop = LDAP_MOD_ADD;
	++attr;
	break;
    default:
	modop = replace ? LDAP_MOD_REPLACE : LDAP_MOD_ADD;
    }

    modop |= LDAP_MOD_BVALUES;

    i = 0;
    if ( pmods != NULL ) {
	for ( ; pmods[ i ] != NULL; ++i ) {
	    if ( strcasecmp( pmods[ i ]->mod_type, attr ) == 0 &&
		    pmods[ i ]->mod_op == modop ) {
		break;
	    }
	}
    }

    if ( pmods == NULL || pmods[ i ] == NULL ) {
	if (( pmods = (LDAPMod **)safe_realloc( pmods, (i + 2) *
		sizeof( LDAPMod * ))) == NULL ) {
	    perror( "safe_realloc" );
	    exit( 1 );
	}
	*pmodsp = pmods;
	pmods[ i + 1 ] = NULL;
	if (( pmods[ i ] = (LDAPMod *)calloc( 1, sizeof( LDAPMod )))
		== NULL ) {
	    perror( "calloc" );
	    exit( 1 );
	}
	pmods[ i ]->mod_op = modop;
	if (( pmods[ i ]->mod_type = strdup( attr )) == NULL ) {
	    perror( "strdup" );
	    exit( 1 );
	}
    }

    if ( value != NULL ) {
	j = 0;
	if ( pmods[ i ]->mod_bvalues != NULL ) {
	    for ( ; pmods[ i ]->mod_bvalues[ j ] != NULL; ++j ) {
		;
	    }
	}
	if (( pmods[ i ]->mod_bvalues =
		(struct berval **)safe_realloc( pmods[ i ]->mod_bvalues,
		(j + 2) * sizeof( struct berval * ))) == NULL ) {
	    perror( "safe_realloc" );
	    exit( 1 );
	}
	pmods[ i ]->mod_bvalues[ j + 1 ] = NULL;
	if (( bvp = (struct berval *)malloc( sizeof( struct berval )))
		== NULL ) {
	    perror( "malloc" );
	    exit( 1 );
	}
	pmods[ i ]->mod_bvalues[ j ] = bvp;

	if ( valsfromfiles && *value == '/' ) {	/* get value from file */
	    if ( fromfile( value, bvp ) < 0 ) {
		exit( 1 );
	    }
	} else {
	    bvp->bv_len = strlen( value );
	    if (( bvp->bv_val = strdup( value )) == NULL ) {
		perror( "strdup" );
		exit( 1 );
	    }
	}
    }
}


domodify( ld, dn, pmods )
    LDAP	*ld;
    char	*dn;
    LDAPMod	**pmods;
{
    int			i, j, k, notascii, op;
    struct berval	*bvp;

    if ( pmods == NULL ) {
	fprintf( stderr, "no attributes to change or add!\n" );
	exit( 1 );
    }

    if ( verbose ) {
	for ( i = 0; pmods[ i ] != NULL; ++i ) {
	    op = pmods[ i ]->mod_op & ~LDAP_MOD_BVALUES;
	    printf( "%s %s:\n", op == LDAP_MOD_REPLACE ?
		    "replace" : op == LDAP_MOD_ADD ?
		    "add" : "delete", pmods[ i ]->mod_type );
	    if ( pmods[ i ]->mod_bvalues != NULL ) {
		for ( j = 0; pmods[ i ]->mod_bvalues[ j ] != NULL; ++j ) {
		    bvp = pmods[ i ]->mod_bvalues[ j ];
		    notascii = 0;
		    for ( k = 0; k < bvp->bv_len; ++k ) {
			if ( !isascii( bvp->bv_val[ k ] )) {
			    notascii = 1;
			    break;
			}
		    }
		    if ( notascii ) {
			printf( "\tNOT ASCII (%d bytes)\n", bvp->bv_len );
		    } else {
			printf( "\t%s\n", bvp->bv_val );
		    }
		}
	    }
	}
    }

    if ( new ) {
	printf( "%sadding new entry %s\n", not ? "!" : "", dn );
    } else {
	printf( "%smodifying entry %s\n", not ? "!" : "", dn );
    }

    if ( !not ) {
	if ( new ) {
	    i = ldap_add_s( ld, dn, pmods );
	} else {
	    i = ldap_modify_s( ld, dn, pmods );
	}
	if ( i != LDAP_SUCCESS ) {
	    ldap_perror( ld, new ? "ldap_add" : "ldap_modify" );
	} else if ( verbose ) {
	    printf( "modify complete\n" );
	}
    } else {
	i = LDAP_SUCCESS;
    }

    return( i );
}


freepmods( pmods )
    LDAPMod	**pmods;
{
    int	i;

    for ( i = 0; pmods[ i ] != NULL; ++i ) {
	if ( pmods[ i ]->mod_bvalues != NULL ) {
	    ber_bvecfree( pmods[ i ]->mod_bvalues );
	}
    }
    free( pmods );
}


int
fromfile( path, bv )
    char		*path;
    struct berval	*bv;
{
	FILE		*fp;
	long		rlen;
	int		eof;

	if (( fp = fopen( path, "r" )) == NULL ) {
	    	perror( path );
		return( -1 );
	}

	if ( fseek( fp, 0L, SEEK_END ) != 0 ) {
		perror( path );
		fclose( fp );
		return( -1 );
	}

	bv->bv_len = ftell( fp );

	if (( bv->bv_val = (char *)malloc( bv->bv_len )) == NULL ) {
		perror( "malloc" );
		fclose( fp );
		return( -1 );
	}

	if ( fseek( fp, 0L, SEEK_SET ) != 0 ) {
		perror( path );
		fclose( fp );
		return( -1 );
	}

	rlen = fread( bv->bv_val, 1, bv->bv_len, fp );
	eof = feof( fp );
	fclose( fp );

	if ( rlen != bv->bv_len ) {
		perror( path );
		free( bv->bv_val );
		return( -1 );
	}

	return( bv->bv_len );
}
