#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "../slap.h"
#include "../back-ldbm/back-ldbm.h"

#define DEFAULT_CONFIGFILE	"%ETCDIR%/slapd.conf"
#define INDEXCMD		"%ETCDIR%/ldif2index"
#define ID2ENTRYCMD		"%ETCDIR%/ldif2id2entry"
#define ID2CHILDRENCMD		"%ETCDIR%/ldif2id2children"
#define MAXARGS      		100

extern void		attr_index_config();
extern char		*str_getline();
extern int		strcasecmp();
extern int		nbackends;
extern Backend		*backends;
extern int		ldap_debug;

int		ldap_debug;
int		ldap_syslog;
int		ldap_syslog_level;
int		global_schemacheck;
int		num_entries_sent;
int		num_bytes_sent;
int		active_threads;
char		*default_referral;
struct objclass	*global_oc;
time_t		currenttime;
pthread_t	listener_tid;
pthread_mutex_t	num_sent_mutex;
pthread_mutex_t	entry2str_mutex;
pthread_mutex_t	active_threads_mutex;
pthread_mutex_t	new_conn_mutex;
pthread_mutex_t	currenttime_mutex;
pthread_mutex_t	replog_mutex;
pthread_mutex_t	ops_mutex;

static void	fork_child();

static char	*indexcmd;
static char	*tailorfile;
static char	*inputfile;
static int      maxkids = 1;
static int      nkids;
 
static void
usage( char *name )
{
	fprintf( stderr, "usage: %s -i inputfile [-d debuglevel] [-f configfile] [-j #jobs]\n", name );
	exit( 1 );
}

main( int argc, char **argv )
{
	int		i, stop, status;
	char		*linep, *buf;
	char		*args[10];
	char		line[BUFSIZ];
	int		lineno, elineno;
	int      	lmax, lcur;
	ID		id;
	Backend		*be = NULL;
	struct berval	bv;
	struct berval	*vals[2];
	Avlnode		*avltypes = NULL;
	extern char	*optarg;

	tailorfile = DEFAULT_CONFIGFILE;
	while ( (i = getopt( argc, argv, "d:f:i:j:" )) != EOF ) {
		switch ( i ) {
		case 'd':	/* turn on debugging */
			ldap_debug = atoi( optarg );
			break;

		case 'f':	/* specify a tailor file */
			tailorfile = strdup( optarg );
			break;

		case 'i':	/* input file */
			inputfile = strdup( optarg );
			break;

		case 'j':	/* number of parallel index procs */
			maxkids = atoi( optarg );
			break;

		default:
			usage( argv[0] );
			break;
		}
	}
	if ( inputfile == NULL ) {
		usage( argv[0] );
	} else {
		if ( freopen( inputfile, "r", stdin ) == NULL ) {
			perror( inputfile );
			exit( 1 );
		}
	}

	/*
	 * initialize stuff and figure out which backend we're dealing with
	 */

	init();
	read_config( tailorfile, &be, NULL );

	for ( i = 0; i < nbackends; i++ ) {
		if ( strcasecmp( be[i].be_type, "ldbm" ) == 0 ) {
			break;
		}
	}
	if ( i == nbackends ) {
		fprintf( stderr, "No ldbm database found in config file\n" );
		exit( 1 );
	}
	be = &backends[i];

	/*
	 * generate the id2entry index
	 */

	args[0] = ID2ENTRYCMD;
	args[1] = "-i";
	args[2] = inputfile;
	args[3] = "-f";
	args[4] = tailorfile;
	if ( ldap_debug ) {
		char	buf[20];

		sprintf( buf, "%d", ldap_debug );
		args[5] = "-d";
		args[6] = buf;
		args[7] = NULL;
	} else {
		args[5] = NULL;
	}
	fork_child( ID2ENTRYCMD, args );

	/*
	 * generate the dn2id and id2children indexes
	 */

	args[0] = ID2CHILDRENCMD;
	args[1] = "-i";
	args[2] = inputfile;
	args[3] = "-f";
	args[4] = tailorfile;
	if ( ldap_debug ) {
		char	buf[20];

		sprintf( buf, "%d", ldap_debug );
		args[5] = "-d";
		args[6] = buf;
		args[7] = NULL;
	} else {
		args[5] = NULL;
	}
	fork_child( ID2CHILDRENCMD, args );

	/*
	 * generate the attribute indexes
	 */

	args[0] = INDEXCMD;
	args[1] = "-i";
	args[2] = inputfile;
	args[3] = "-f";
	args[4] = tailorfile;
	if ( ldap_debug ) {
		char	buf[20];

		sprintf( buf, "%d", ldap_debug );
		args[5] = "-d";
		args[6] = buf;
		args[7] = NULL;		/* will hold the attribute name */
		args[8] = NULL;
	} else {
		args[5] = NULL;		/* will hold the attribute name */
		args[6] = NULL;
	}
	id = 0;
	stop = 0;
	buf = NULL;
	lineno = 0;
	lcur = lmax = 0;
	vals[0] = &bv;
	vals[1] = NULL;
	while ( ! stop ) {
		char		*type, *val, *s;
		int		vlen, indexmask, syntaxmask;
		Datum		key, data;

		if ( fgets( line, sizeof(line), stdin ) != NULL ) {
			int     len;

			lineno++;
			len = strlen( line );
			while ( lcur + len + 1 > lmax ) {
				lmax += BUFSIZ;
				buf = (char *) ch_realloc( buf, lmax );
			}
			strcpy( buf + lcur, line );
			lcur += len;
		} else {
			stop = 1;
		}
		if ( line[0] == '\n' || stop && buf && *buf ) {
			id++;
			s = buf;
			elineno = 0;
			while ( (linep = str_getline( &s )) != NULL ) {
				elineno++;
				if ( str_parse_line( linep, &type, &val, &vlen )
				    != 0 ) {
					Debug( LDAP_DEBUG_PARSE,
			    "bad line %d in entry ending at line %d ignored\n",
					    elineno, lineno, 0 );
					continue;
				}

				if ( !isascii( *type ) || isdigit( *type ) )
					continue;

				type = strdup( type );
				if ( avl_insert( &avltypes, type, strcasecmp,
				    avl_dup_error ) != 0 ) {
					free( type );
				} else {
					attr_masks( be->be_private, type,
					    &indexmask, &syntaxmask );
					if ( indexmask ) {
						args[ldap_debug ? 7 : 5] = type;
						fork_child( INDEXCMD, args );
					}
				}
			}
			*buf = '\0';
			lcur = 0;
		}
	}
	(*be->be_close)( be );

	while ( nkids > 0 ) {
		wait( &status );
		nkids--;
	}

	exit( 0 );
}

static void
fork_child( char *prog, char *args[] )
{
	int	status, pid;

	while ( nkids >= maxkids ) {
		wait( &status );
		nkids--;
	}

	switch ( pid = fork() ) {
	case 0:		/* child */
		execvp( prog, args );
		fprintf( stderr, "%s: ", prog );
		perror( "execv" );
		exit( -1 );
		break;

	case -1:	/* trouble */
		fprintf( stderr, "Could not fork to run %s\n", prog );
		perror( "fork" );
		break;

	default:	/* parent */
		nkids++;
		break;
	}
}
