/*
 * Copyright (c) 1990, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <errno.h>
#include <stdio.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include "portable.h"
#include "slurp.h"

/*
 * Externs
 */
extern FILE *lock_fopen( char *, char *, FILE ** );
extern char *ch_malloc( unsigned long );
#ifndef SYSERRLIST_IN_STDIO
extern char *sys_errlist[];
#endif

/*
 * Forward declarations
 */
static int duplicate_replog( char *, char * );




/*
 * Copy the replication log.  Returns 0 on success, 1 if a temporary
 * error occurs, and -1 if a fatal error occurs.
 */
int
copy_replog( char *replogfile, char *myreplogfile )
{
    int		rc = 0;
    FILE	*rfp;	/* replog fp */
    FILE	*lfp;	/* replog lockfile fp */
    FILE	*dfp;	/* duplicate replog fp */
    static char	buf[ MAXPATHLEN ];
    char	*p;

    Debug( LDAP_DEBUG_ARGS,
	    "Trace: copy_replog: replogfile = %s, myreplogfile = %s\n", 
	    replogfile, myreplogfile, 0 );

    /*
     * Make sure the destination directory is writable.  If not, exit
     * with a fatal error.
     */
    strcpy( buf, replogfile );
    if (( p = strrchr( buf, '/' )) == NULL ) {
	strcpy( buf, "." );
    } else {
	*p = '\0';
    }
    if ( access( buf, W_OK ) < 0 ) {
	Debug( LDAP_DEBUG_ANY,
		"Error: copy_replog (%d): Directory %s is not writable\n",
		getpid(), buf, 0 );
	return( -1 );
    }
    strcpy( buf, myreplogfile );
    if (( p = strrchr( buf, '/' )) == NULL ) {
	strcpy( buf, "." );
    } else {
	*p = '\0';
    }
    if ( access( buf, W_OK ) < 0 ) {
	Debug( LDAP_DEBUG_ANY,
		"Error: copy_replog (%d): Directory %s is not writable\n",
		getpid(), buf, 0 );
	return( -1 );
    }

    /* lock replogfile */
    Debug( LDAP_DEBUG_ARGS,
	    "Trace: copy_replog (%d): about to lock \"%s\"\n",
	    getpid(), replogfile, 0 );
    rfp = lock_fopen( replogfile, "r", &lfp );
    if ( rfp == NULL ) {
	Debug( LDAP_DEBUG_ANY,
		"Error: copy_replog (%d): Couldn't lock replog \"%s\": %s\n",
		getpid(), replogfile, sys_errlist[ errno ] );
	return( 1 );
    }

    /*
     * If the destination file is nonempty, then there is some
     * unfinished business there, and we need to preserve the
     * contents.
     */
    if ( file_nonempty( myreplogfile )) {
	Debug( LDAP_DEBUG_ARGS,
		"Trace: copy_replog (%d): dest file nonempty\n",
		getpid(), 0, 0 );
	rc = duplicate_replog( replogfile, myreplogfile );
	if ( rc == 0 ) {
	    /* truncate replication log */
	    truncate( replogfile, (off_t) 0 );
	}
    } else {
	/*
	 * Optimization: 
	 * We know the destination file is empty or nonexistent.
	 * Try to rename(2) the file.  If it succeeds, great.  If not,
	 * then if the errno was EXDEV, then replogfile and myreplogfile
	 * are on different filesystems, and we need to copy the file.
	 * If the error wasn't EXDEV, then, well, give up.
	 */
	if (( rc = rename( replogfile, myreplogfile )) < 0 ) {
	    if ( errno == EXDEV ) {
		/* Files are on different file systems - duplicate */
		rc = duplicate_replog( replogfile, myreplogfile );
		Debug( LDAP_DEBUG_ARGS,
			"Trace: copy_replog (%d): duped file\n",
			getpid(), 0, 0 );
		if ( rc == 0 ) {
		    /* truncate replication log */
		    truncate( replogfile, (off_t) 0 );
		}
	    } else {
		/* rename() failed for some other reason */
		Debug( LDAP_DEBUG_ANY,
		    "Error: copy_replog: Cannot rename \"%s\" as \"%s\": %s\n",
		    replogfile, myreplogfile, sys_errlist[ errno ] );
		rc = -1;
	    }
	} else {
	    /*
	     * The rename operation succeeded.  It doesn't matter if we
	     * don't recreate the file, since slapd will re-create it.
	     */
	    Debug( LDAP_DEBUG_ARGS,
		    "Trace: copy_replog (%d): mved file\n",
		    getpid(), 0, 0 );
	    rc = 0;
	}
    }

    if ( lock_fclose( rfp, lfp ) == EOF ) {
	Debug( LDAP_DEBUG_ANY,
		"Error: copy_replog (%d): Error closing \"%s\"\n",
		getpid(), replogfile, 0 );
    }
    sprintf( buf, "%s.lock", replogfile );
    unlink( buf );
    return( rc );
}





/*
 * Copy "replogfile" to "myreplogfile".  Contents of "myreplogfile" are
 * preseved (data are appended).  Returns 0 on sucess, -1 on failure.
 */
static int
duplicate_replog( char *replogfile, char *myreplogfile )
{
    int		sfd;	/* source file */
    int		dfd;	/* destination file */
    struct stat	ssbuf;
    struct stat	dsbuf;
    long	bsize, sbsize, dbsize;
    char	*buf;

    /* open files */
    Debug( LDAP_DEBUG_ARGS,
	    "Trace: duplicate_replog (%d) : open \"%s\" for read\n",
	    getpid(), replogfile, 0 );
    if (( sfd = open( replogfile, O_RDONLY )) < 0 ) {
	Debug( LDAP_DEBUG_ANY,
	    "Error: duplicate_replog (%d): Cannot open \"%s\" for read: %s\n",
	    getpid(), replogfile, sys_errlist[ errno ] );
	return( -1 );
    }

    /* if "myreplogfile" doesn't exist, create, else open for append */
    Debug( LDAP_DEBUG_ARGS,
	    "Trace: duplicate_replog (%d): open \"%s\" for write\n",
	    getpid(), myreplogfile, 0 );
    if (( dfd = open( myreplogfile, O_RDWR | O_APPEND | O_CREAT,
	    S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP )) < 0 ) {
	 Debug( LDAP_DEBUG_ANY,
	    "Error: duplicate_replog (%d): Cannot open \"%s\" for write: %s\n",
	    getpid(), myreplogfile, sys_errlist[ errno ] );
	return( -1 );
    }

    /* find best block size for copy */
    if ( fstat( sfd, &ssbuf ) < 0 ) {
	sbsize = 512;
    } else {
	sbsize = ssbuf.st_blksize;
    }
    if ( fstat( dfd, &dsbuf ) < 0 ) {
	dbsize = 512;
    } else {
	dbsize = dsbuf.st_blksize;
    }
    bsize = ( sbsize < dbsize ) ? sbsize : dbsize;
    buf = ch_malloc( bsize );

    for (;;) {
	int	nb;

	nb = read( sfd, buf, bsize );
	if ( nb < 0 ) {
	    Debug( LDAP_DEBUG_ANY,
		"Error: duplicate_replog (%d): read from \"%s\" failed: %s\n",
		getpid(), replogfile, sys_errlist[ errno ] );
	} else {
	    write( dfd, buf, nb );
	}
	if ( nb != bsize ) {
	    break;
	}
    }
    free( buf );
    close( sfd );
    close( dfd );
    return( 0 );
}
