/*
 * Copyright (c) 1990, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <stdio.h>
#include <syslog.h>
#include <malloc.h>
#include <string.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <signal.h>

#include "portable.h"
#include "slurp.h"
#include "../slapd/slap.h"

/* Externs */
#ifndef SYSERRLIST_IN_STDIO
extern char *sys_errlist[];
#endif
extern replicainfo *replicas;
extern char *slurpd_rdir;

/* Globals */
/* linked list of replication children */
struct repl_instance	*repl_instances;

/* Extern functions */
extern FILE *lock_fopen( char *, char *, FILE ** );
extern int lock_fclose( FILE *, FILE * );
extern struct repl_s *get_replog_entry_s( FILE * );
extern int getmaxopenfiles();
extern int replicate( replicainfo * );

/* Forward references */
static void add_replication_records( struct repl_s * );
static pid_t child_running( char * );
static void start_child( char * );
static void add_child_instance( pid_t, char * );
static struct replicainfo *get_replicainfo( char  * );
static void usr1_handler();

/* Static variables */
static char	prevtime[ LINE_WIDTH ];




/*
 * Read through "myreplogfile" and distribute any work to appropriate
 * replication processes.  Restarts child if child's linger time has
 * expired and child has died.
 */
void
distribute_work(
char *myreplogfile )
{
    FILE		*fp;
    FILE		*lfp;
    struct repl_s	*srepl;
    char		fname[ MAXPATHLEN ];
    int			rc;

    /* lock myreplogfile */
    fp = lock_fopen( myreplogfile, "r", &lfp );
    if ( fp == NULL ) {
	Debug( LDAP_DEBUG_ANY,
	    "Error: distribute_work (%d): Can't lock replog \"%s\": %s\n",
	    getpid(), myreplogfile, sys_errlist[ errno ] );
	return;
    }
    prevtime[ 0 ] = '\0';

    /*
     * Read replication records from "myreplogfile", add them to appropriate
     * per-replica files.
     */
    for ( srepl = get_replog_entry_s( fp );
	    srepl != NULL; srepl = get_replog_entry_s( fp )) {
	add_replication_records( srepl );
	free_repl_s( srepl );
    }

    /*
     * At this point, work has been distributed to the individual
     * files for each replica.  Truncate and unlock our copy of
     * the replication logfile.
     */
    Debug( LDAP_DEBUG_ARGS,
	    "Trace: distribute_work (%d): About to truncate \"%s\"\n",
	    getpid(), myreplogfile, 0 );
    rc = truncate( myreplogfile, 0L );
    if ( rc != 0 ) {
	Debug( LDAP_DEBUG_ANY,
		"Error: distribute_work (%d): Truncate of \"%s\" failed:  %s\n",
		getpid(), myreplogfile, sys_errlist[ errno ] );
    }
    lock_fclose( fp, lfp );
    sprintf( fname, "%s.lock", myreplogfile );
    unlink( fname );
}



/*
 * Given a replication record in simple format, append it to the
 * replica file for each replica specified.
 */
static void
add_replication_records( struct repl_s *srepl )
{
    int		i;
    char	fname[ MAXPATHLEN ];
    FILE	**rfps;
    FILE	**lfps;
    char	**files;
    int		done;
    char	*p, *t;
    char	buf[ LINE_WIDTH + 1 ], tbuf[ LINE_WIDTH + 1 ];
    char	*type, *value;
    int		uniqifier;
    int		len;
    pid_t	pid;

    if ( srepl == NULL ) {
	Debug( LDAP_DEBUG_ANY,
		"Error: add_replication_record (%d): no replicas given\n",
		getpid(), 0, 0 );
	return;
    }
    
    rfps = (FILE **) malloc( sizeof( FILE * ) * srepl->nreplicas );
    lfps = (FILE **) malloc( sizeof( FILE * ) * srepl->nreplicas );
    files = (char **) malloc( sizeof( char * ) * srepl->nreplicas );
    for ( i = 0; i < srepl->nreplicas; i++ ) {
	/* Construct replication file names and open files */
	sprintf( fname, "%s/%s.replog", slurpd_rdir, srepl->replicas[ i ] );
	files[ i ] = strdup( fname );
	rfps[ i ] = lock_fopen( fname, "a", &( lfps[ i ] ));
	if ( rfps[ i ] == NULL ) {
	    Debug( LDAP_DEBUG_ANY,
		"Error: add_replication_records (%d): Can't lock \"%s\": %s\n",
		getpid(), fname, sys_errlist[ errno ] );
	    continue;
	}
	fprintf( rfps[ i ], "replica: %s\n", srepl->replicas[ i ] );
    }

    /* Copy out replication data to each file */
    done = 0;
    uniqifier = 0;
    p = srepl->replbuf;

    /* uniqify time by appending a 4-digit monotonically-increasing number */
    if ( !strncasecmp( p, "time:", 5 )) {
	t = tbuf;
	for ( t = tbuf; *p != '\n' && *p != '\0'; *t++ = *p++ ); /* copy */
	*t = '\0';
	(void)str_parse_line( tbuf, &type, &value, &len );
	if ( strcmp( value, prevtime )) {
	    uniqifier = 0;
	} else {
	    uniqifier++;
	}
	strcpy( prevtime, value );
	sprintf( buf, "time: %s.%04d\n", value, uniqifier );
	for ( i = 0; i < srepl->nreplicas; i++ ) {
	    fputs( buf, rfps[ i ]);
	}
	for ( ; *p != '\n'; p++ );	/* skip rest of line */
    }

    if ( *p != '\0' ) {
	p++; /* skip newline */
	for ( i = 0; i < srepl->nreplicas; i++ ) {
	    fputs( p, rfps[ i ]);	/* write rest of replication record */
	    fputs( "\n", rfps[ i ]);
	}
    }

    /* Close files and start children to service each replica */
    for ( i = 0; i < srepl->nreplicas; i++ ) {
	fflush( rfps[ i ]);
	lock_fclose( rfps[ i ], lfps[ i ] );
	sprintf( fname, "%s.lock", files[ i ] );
	unlink( fname );
	free( files[ i ]);
	if (( pid = child_running( srepl->replicas[ i ] )) == 0 ) {
	    start_child( srepl->replicas[ i ] );
	} else {
	    Debug( LDAP_DEBUG_ARGS,
		"Trace: add_replication_records (%d): child %s exists\n",
		getpid(), srepl->replicas[ i ], 0 );
	    kill( pid, SIGUSR1 ); /* USR1 means "wake up, work to do" */
	}
    }
    free( rfps );
    free( lfps );
    free( files );
}



/*
 * Return the pid of the child process serving "replica".  If no child
 * is running for the replica, return 0.  It is assumed that "replica" is
 * given as a "hostname:port" pair.
 */
static pid_t
child_running( char *replica )
{
    replicainfo *r;

    if (( r = get_replicainfo( replica )) != NULL ) {
	return( r->repl_pid );
    }
    return (pid_t) 0;
}




/*
 * Fork a child process to service "replica" and add to the list
 * of children.  It is assumed that "replica" is given as a
 * "hostname:port" pair.
 */
static void
start_child( char *replica  )
{
    int		rc;
    pid_t	pid;
    int		i;
    replicainfo	*r;

    if ( replica == NULL || *replica == '\0' ) {
	Debug( LDAP_DEBUG_ANY,
		"Error: start_child (%d): null or empty replica name\n",
		getpid(), 0, 0 );
	return;
    }
    switch ( pid = fork()) {
    case -1:	/* error */
	Debug( LDAP_DEBUG_ANY, "Error: cannot fork: %s\n",
		sys_errlist[ errno ], 0, 0 );
	return;
    case 0:	/* child */
	for ( i = 3; i < getmaxopenfiles(); i++ ) {
	    close( i );
	}
	if (( r = get_replicainfo( replica )) == NULL ) {
	    Debug( LDAP_DEBUG_ANY,
		    "Error: start_child (%d): no replicainfo found for %s\n",
		    replica, 0, 0 );
	    exit( -1 );
	} else {
	    exit( replicate( r ));
	}
    default:	/* parent */
	add_child_instance( pid, replica );
	Debug( LDAP_DEBUG_ARGS,
		"Trace: start_child (%d): Start child for \"%s\" , pid %d\n",
		getpid(), replica, pid );
	return;
    }
}



/*
 * Clean up exiting/dying child processes, and remove them from the
 * list of replica processes.
 */
void
reap_child()
{
    WAITSTATUSTYPE	status;
    pid_t		pid;
    replicainfo		*r;

#ifdef USE_WAITPID
    while((pid = waitpid( (pid_t) -1, &status, WAIT_FLAGS )) > 0 ) {
#else
    while (( pid = wait3( &status, WNOHANG, 0 )) > 0 ) {
#endif
	if ( WIFEXITED( status )) {
	    if ( WEXITSTATUS( status )) {
		Debug( LDAP_DEBUG_ANY,
			"Error: child %d exited with %d\n", pid,
			WEXITSTATUS( status ), 0 );
	    } else {
		Debug( LDAP_DEBUG_ARGS, "Trace: child %d done\n", pid, 0, 0 );
	    }
	} else {
	    if ( WIFSIGNALED( status )) {
		Debug( LDAP_DEBUG_ANY, "Error: child %d killed with %d\n", pid,
			WTERMSIG( status ), 0 );
	    } else {
		Debug( LDAP_DEBUG_ANY, "Error: child %d died\n", pid, 0, 0 );
	    }
	}
	/* Set replica's pid to 0, and ldap ptr to NULL */
	r = replicas;
	while ( r != NULL ) {
	    if ( r->repl_pid == pid ) {
		r->repl_pid = (pid_t) 0;
		r->repl_ldp = NULL;
	    } else {
		r = r->repl_next;
	    }
	}
    }
}



/* 
 * Update the pid for a replica instance.  It is assumed that "replica" is
 * given as a "hostname:port" pair.
 */
static void
add_child_instance(
pid_t pid,
char *replica )
{
    replicainfo	*r;

    if (( r = get_replicainfo( replica )) != NULL ) {
	r->repl_pid = pid;
    } else {
	Debug( LDAP_DEBUG_ANY,
		"Error: add_child_instance: no replicainfo found for host %s\n",
		replica, 0, 0 );
    }
    return;
}




/*
 * Traverse the linked list of replicainfo structs and return the one
 * whose host name and port match "replica".  "replica" must be given as
 * a "hostname:port" pair.
 */
static struct replicainfo *
get_replicainfo( 
char	*replica )
{
    replicainfo	*r;
    char	buf[ REPLBUFLEN ];
    int		port;
    char	*p;

    r = replicas;
    while ( r != NULL ) {

	if ( !strncasecmp( r->repl_hostname, replica,
		strlen( r->repl_hostname ))) {
	    if (( p = strchr( replica, ':' )) == NULL ) {
		port = LDAP_PORT;
	    } else {
		p++;
		port = atoi( p );
		if ( port == 0 ) {
		    port = LDAP_PORT;
		}
	    }
	    if ( r->repl_port == port ) {
		return( r );
	    }
	}
	r = r->repl_next;
    }
    Debug( LDAP_DEBUG_ANY,
	    "Error: get_replicainfo: no replicainfo found for host %s\n",
	    replica, 0, 0 );
    return( NULL );
}
