#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <lber.h>
#include <ldap.h>

#include "portable.h"
#include "slurp.h"
#include "../slapd/slap.h"

/* Externs */
#ifndef SYSERRLIST_IN_STDIO
extern char *sys_errlist[];
#endif

/* Extern functions */
extern char *ch_malloc( unsigned long );
extern char *ch_realloc( char *, unsigned long );
extern struct timeval *str2optime( char * );
extern char *getlinecont( FILE * );
extern char *str_getline( char ** );

/* Forward references */
static char 	**get_repl_hosts( char *, int *, char ** );
static int	gettype( char * );
static int	getchangetype( char *);
struct repl	*get_replog_entry( FILE *, int * );
void		free_repl( struct repl * );
static int	cmp_timeval( struct timeval *, struct timeval * );
static char	*get_record( FILE * );
static char *	skip_replica_lines( char * );


/*
 * Return the replication record in the file referenced by rfp.
 * Return in "simple" format - char ** for each replica name, a
 * count of the number of replicas, and an offset pointing to the
 * first character past the "replica:" information.
 */
struct repl_s *
get_replog_entry_s(
FILE *rfp )
{
    struct repl_s	*srepl;
    char		*rp;
    char		*replbuf, *tmp;

    if (( replbuf = get_record( rfp )) == NULL ) {
	return( NULL );
    }

    srepl = ( struct repl_s * ) ch_malloc( sizeof( struct repl_s ));
    tmp = strdup(replbuf);
    srepl->replicas = get_repl_hosts( tmp, &(srepl->nreplicas), &rp );
    srepl->replbuf = strdup( skip_replica_lines( replbuf ));
    free( tmp );
    free( replbuf );
    return( srepl );
}






/*
 * Return the replication record in the file referenced by rfp.
 */
#define	BEGIN		0
#define	GOT_DN		1
#define	GOT_TIME	2
#define	GOT_CHANGETYPE	4
#define	GOT_ALL		( GOT_DN | GOT_TIME | GOT_CHANGETYPE )

struct repl *
get_replog_entry(
FILE *rfp, 
int *status )
{
    struct repl		*rpl;
    int			state;
    char		*buf, *rbuf, *replbuf, *rp;
    long		buflen;
    char		*type, *value;
    int			len;
    int			nml;


    *status = 0;

    if (( replbuf = get_record( rfp )) == NULL ) {
	*status |= NO_MORE_WORK;
	return( NULL );
    }

    state = BEGIN;
    nml = 0;	/* number of modification information entries */
    rp = replbuf;

    rpl = ( struct repl * ) ch_malloc( sizeof( struct repl ));
    rpl->replicas = get_repl_hosts( replbuf, &(rpl->nreplicas), &rp );
    rpl->mods = NULL;

    for (;;) {
	if (( state == GOT_ALL ) || ( buf = str_getline( &rp )) == NULL ) {
	    break;
	}
	/*
	 * If we're processing a rejection log, then the first line
	 * of each replication record will begin with "ERROR" - just
	 * ignore it.
	 */
	if ( strncmp( buf, ERROR_STR, strlen( ERROR_STR )) == 0 ) {
	    continue;
	}
	buflen = ( long ) strlen( buf );
	if ( str_parse_line( buf, &type, &value, &len ) < 0 ) {
	    Debug( LDAP_DEBUG_ANY,
		    "Error: malformed replog file (1)(%d), buflen = %ld\n",
		    getpid(), buflen, 0 );
	    return( NULL );
	}
	switch ( gettype( type )) {
	case T_CHANGETYPE:
	    rpl->changetype = getchangetype( value );
	    state |= GOT_CHANGETYPE;
	    break;
	case T_TIME:
	    rpl->time = strdup( value );
	    state |= GOT_TIME;
	    break;
	case T_DN:
	    rpl->dn = strdup( value );
	    state |= GOT_DN;
	    break;
	default:
	    if ( !( state == GOT_ALL )) {
		Debug( LDAP_DEBUG_ANY,
			"Error: get_replog_entry (%d): bad type <%s>\n",
			getpid(), type, 0 );
		free_repl( rpl );
		return( NULL );
	    }
	}
    }

    if ( state != GOT_ALL ) {
	Debug( LDAP_DEBUG_ANY,
		"Error: get_replog_entry (%d): malformed replog file\n",
		getpid(), 0, 0 );
	free_repl( rpl );
	return( NULL );
    }

    for (;;) {
	if (( buf = str_getline( &rp )) == NULL ) {
	    break;
	}
	buflen = ( long ) strlen( buf );
	if (( buflen == 1 ) && ( buf[ 0 ] == '-' )) {
	    type = strdup( "-" );
	    value = NULL;
	} else {
	    if ( str_parse_line( buf, &type, &value, &len ) < 0 ) {
		Debug( LDAP_DEBUG_ANY,
			"Error (%d): malformed replog line \"%s\"\n",
			getpid(), buf, 0 );
		return( NULL );
	    }
	}
	rpl->mods = (struct modinfo *) ch_realloc( (char *) rpl->mods,
	    sizeof( struct modinfo ) * ( nml + 2 ));
	rpl->mods[ nml ].type = strdup( type );
	if ( value != NULL ) {
	    rpl->mods[ nml ].value = strdup( value );
	} else {
	    rpl->mods[ nml ].value = NULL;
	}
	rpl->mods[ nml ].len = len;
	rpl->mods[ nml + 1 ].type = NULL;
	rpl->mods[ nml + 1 ].value = NULL;
	nml++;
    }

    free( replbuf );
    return( rpl );
}




/*
 * Return the first replication record whose time value is greater
 * than "lastoptime".  Note - does not reposition the offset into
 * the file pointed to by rfp.  It is the caller's responsibility
 * to keep the file offset correct.  Leaves the file pointer pointing
 * the the first replication  record following the one returned.
 */
struct repl *
get_replog_entry_t(
FILE *rfp,
struct timeval *lastoptime,
int *status )
{
    struct timeval	tv, *tvp;
    struct repl		*rpl;

    for (;;) {
	rpl = get_replog_entry( rfp, status );
	if ( rpl == NULL ) {
	    return( NULL );
	}
	tvp = str2optime( rpl->time );
	if ( cmp_timeval( tvp, lastoptime )) {
	    return( rpl );
	} else {
	    free_repl( rpl );
	    return( NULL );
	}
    }
}




/*
 * Free a repl_s structure.
 */
void
free_repl_s( struct repl_s *p )
{
    int	i;

    if ( p != NULL ) {
	if ( p->replicas != NULL ) {
	    for ( i = 0; p->replicas[ i ] != NULL; i++ ) {
		free( p->replicas[ i ] );
	    }
	    free( p->replicas );
	}
	if ( p->replbuf != NULL ) {
	    free( p->replbuf );
	}
	free( p );
    }
}




/*
 * Free a repl structure.
 */
void
free_repl( struct repl *p )
{
    int	i;

    if ( p != NULL ) {
	if ( p->replicas != NULL ) {
	    for ( i = 0; p->replicas[ i ] != NULL; i++ ) {
		free( p->replicas[ i ] );
	    }
	    free( p->replicas );
	}
	if ( p->time != NULL ) {
	    free( p->time );
	}
	if ( p->dn != NULL ) {
	    free( p->dn );
	}
	if ( p->mods != NULL ) {
	    for ( i = 0; p->mods[ i ].type != NULL; i++ ) {
		free( p->mods[ i ].type );
		if ( p->mods[ i ].value != NULL ) {
		    free( p->mods[ i ].value );
		}
	    }
	    free( p->mods );
	}
	free( p );
    }
}




static char **
get_repl_hosts( char *replbuf, int *r_nreplicas, char **r_rp )
{
    char			buf[ LINE_WIDTH + 1 ];
    char			*type, *value, *line, **r_replicas;
    static char 		*replicas[ MAXREPLICAS ];
    int				nreplicas, len;

    if ( replbuf == NULL ) {
	return( NULL );
    }

    nreplicas = 0;

    /*
     * Get the host names of the replicas
     */
    *r_rp = replbuf;
    for (;;) {
	/* If this is a reject log, we need to skip over the ERROR: line */
	if ( !strncmp( *r_rp, ERROR_STR, strlen( ERROR_STR ))) {
	    line = str_getline( r_rp );
	    if ( line == NULL ) {
		break;
	    }
	}
	if ( strncasecmp( *r_rp, "replica:", 7 )) {
	    break;
	}
	line = str_getline( r_rp );
	if ( line == NULL ) {
	    break;
	}
	if ( str_parse_line( line, &type, &value, &len ) < 0 ) {
	    return( NULL );
	}
	replicas[ nreplicas ] = strdup( value );
	nreplicas++;
    }

    if ( nreplicas == 0 ) {
	Debug( LDAP_DEBUG_ANY,
		"Error: get_repl_hosts (%d): bad replog file - no replicas \n",
		getpid(), 0, 0 );
	return( NULL );
    }

    replicas[ nreplicas ] = NULL;
    r_replicas = (char ** ) 
	    ch_malloc( sizeof( char * ) * ( nreplicas + 1 ));
    memcpy( r_replicas, replicas,  sizeof( char * ) * ( nreplicas + 1));
    *r_nreplicas = nreplicas;

    return( r_replicas );
}



/*
 * Return nonzero value if the file size is greater than a predetermined
 * size and we're more than 3/4 of the way through the file.
 */
int
time_to_trim(
FILE *fp )
{
    struct stat	stbuf;
    int		fd;
    off_t	pos, qsize;

    fd = fileno( fp );
    if ( fstat( fd, &stbuf ) < 0 ) {
	Debug( LDAP_DEBUG_ANY,
		"Error (%d): Can't fstat file descriptor %d: %s\n",
		getpid(), fd, sys_errlist[ errno ] );
	return( 0 );
    }
    
    pos = ftell( fp );
    qsize = stbuf.st_size >> 2;		/* 1/4 size of the file */
    if (( stbuf.st_size > MIN_TRIM_FILESIZE ) &&
	    ( pos > ( stbuf.st_size - qsize ))) {
	return( 1 );
    } else {
	return( 0 );
    }
}



static int
gettype(
char *type)
{
    if ( !strcmp( type, T_CHANGETYPESTR )) {
	return( T_CHANGETYPE );
    }
    if ( !strcmp( type, T_TIMESTR )) {
	return( T_TIME );
    }
    if ( !strcmp( type, T_DNSTR )) {
	return( T_DN );
    }
    return( T_ERR );
}



static int
getchangetype(
char *type)
{
    if ( !strcmp( type, T_ADDCTSTR )) {
	return( T_ADDCT );
    }
    if ( !strcmp( type, T_MODIFYCTSTR )) {
	return( T_MODIFYCT );
    }
    if ( !strcmp( type, T_DELETECTSTR )) {
	return( T_DELETECT );
    }
    if ( !strcmp( type, T_MODRDNCTSTR )) {
	return( T_MODRDNCT );
    }
    return( T_ERR );
}




/*
 * Return 1 if time1 > time2, -1 if time1 < time2, 0 if they are equal
 */
static int
cmp_timeval(
struct timeval *t1,
struct timeval *t2 )
{
    if ( t1->tv_sec > t2->tv_sec ) {
	return( 1 );
    } else if ( t2->tv_sec > t1->tv_sec ) {
	return( -1 );
    } else { /* seconds are equal */
	if ( t1->tv_usec > t2->tv_usec ) {
	    return( 1 );
	} else if ( t2->tv_usec > t1->tv_usec ) {
	    return( -1 );
        } else {
	    return( 0 );
	}
    }
}



/*
 * Get the next "record" from the file pointed to by fp.  A "record"
 * is delimited by two consecutive newlines.
 */
static char *
get_record(
FILE *fp )
{
    int		len;
    static char	line[BUFSIZ];
    char	*buf = NULL;
    static int	lcur, lmax;

    lcur = lmax = 0;

    while (( fgets( line, sizeof(line), fp ) != NULL ) &&
	    (( len = strlen( line )) > 1 )) {
	while ( lcur + len + 1 > lmax ) {
	    lmax += BUFSIZ;
	    buf = (char *) ch_realloc( buf, lmax );
	}
	strcpy( buf + lcur, line );
	lcur += len;
    }
    return( buf );
}


/*
 * Find the first line which is not a "replica:" line in buf.
 * Returns a pointer to the line.  Returns NULL if there are
 * only "replica:" lines in buf.
 */
static char *
skip_replica_lines(
char *buf )
{
    char *p = buf;
    for (;;) {
	if ( strncasecmp( p, "replica:", 8 )) {
	    return( p );
	}
	while (( *p != '\0' )  && ( *p != '\n' )) {
	    p++;
	}
	if ( *p == '\0' ) {
	    return ( NULL );
	} else {
	    p++;
	}
    }
}

