/*
 * Copyright (c) 1990, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>

#ifdef KERBEROS
#include <krb.h>
#endif /* KERBEROS */

#include <lber.h>
#include <ldap.h>

#include "portable.h"
#include "slurp.h"

/* Forward references */
static int get_changetype( char * );
static struct berval **make_singlevalued_berval( char	*, int );
static int op_ldap_add( struct replicainfo *, struct repl *);
static int op_ldap_modify( struct replicainfo *, struct repl * );
static int op_ldap_delete( struct replicainfo *, struct repl * );
static int op_ldap_modrdn( struct replicainfo *, struct repl * );
static LDAPMod *alloc_ldapmod();
static void free_ldapmod();
static void free_ldmarr( LDAPMod ** );
static int getmodtype( char * );
static void dump_ldm_array( LDAPMod ** );
static char **read_krbnames( replicainfo * );
void upcase( char * );


/* External references */
#ifndef SYSERRLIST_IN_STDIO
extern char *sys_errlist[];
#endif
extern char *ch_malloc( unsigned long );

static char *kattrs[] = {"kerberosName", NULL };
static struct timeval kst = {30L, 0L};

/*
 * Determine the type of ldap operation being performed and call the
 * appropriate routine.
 */
int
perform_ldap_operation(
replicainfo *replica,
struct repl *rpl )
{
    int rc;

    Debug( LDAP_DEBUG_ARGS,
	    "Trace: (%d) perform_ldap_operation: begin\n", getpid(), 0, 0 );
    if ( replica->repl_ldp == NULL ) {
	rc = do_bind( replica );
	if ( rc != LDAP_SUCCESS ) {
	    return( rc );
	}
    }

    switch ( rpl->changetype ) {
    case T_ADDCT:
	rc = op_ldap_add( replica, rpl );
	break;
    case T_MODIFYCT:
	rc = op_ldap_modify( replica, rpl );
	break;
    case T_DELETECT:
	rc = op_ldap_delete( replica, rpl );
	break;
    case T_MODRDNCT:
	rc = op_ldap_modrdn( replica, rpl );
	break;
    default:
	Debug( LDAP_DEBUG_ANY,
		"Error: (%d) perform_ldap_operation: error bad op \"%s\"\n",
		getpid(), rpl->changetype, 0 );
    }
    return( rc );
}


/*
 * Perform an ldap add operation.
 */
static int
op_ldap_add(
replicainfo *replica,
struct repl *rpl )
{
    modinfo	*p;
    int		nattrs, rc = 0, i;
    LDAPMod	*ldm, **ldmarr;

    nattrs = i = 0;
    ldmarr = NULL;

    /*
     * Construct a null-terminated array of LDAPMod structs.
     */
    p = rpl->mods;
    while ( p[ i ].type != NULL ) {
	ldm = alloc_ldapmod();
	ldmarr = ( LDAPMod ** ) ch_realloc( ldmarr,
		( nattrs + 2 ) * sizeof( LDAPMod * ));
	ldmarr[ nattrs ] = ldm;
	ldm->mod_op = LDAP_MOD_BVALUES;
	ldm->mod_type = p[ i ].type;
	ldm->mod_bvalues = make_singlevalued_berval( p[ i ].value, p[ i ].len );
	i++;
	nattrs++;
    }

    if ( ldmarr != NULL ) {
	Debug( LDAP_DEBUG_ARGS, "Trace: (%d) ldap_add_s( ldp, %s, ldmarr)\n",
		getpid(), rpl->dn, 0 );
	ldmarr[ nattrs ] = NULL;

	rc = ldap_add_s( replica->repl_ldp, rpl->dn, ldmarr );

	if ( rc != LDAP_SUCCESS ) {
	    Debug( LDAP_DEBUG_ANY, "Error: (%d) ldap_add_s failed: %s\n",
		    getpid(), ldap_err2string( rc ), 0 );
	}
    } else {
	Debug( LDAP_DEBUG_ARGS,
		"Trace: (%d) op_ldap_add: no mods to do!", getpid(), 0, 0 );
    }
    free_ldmarr( ldmarr );
    return( rc );
}




/*
 * Perform an ldap modify operation.
 */
#define	AWAITING_OP -1
static int
op_ldap_modify(
replicainfo *replica,
struct repl *rpl )
{
    modinfo	*p;
    int		state;	/* This code is a simple-minded state machine */
    int		nvals;	/* Number of values we're modifying */
    int		nops;	/* Number of LDAPMod structs in ldmarr */
    LDAPMod	*ldm, *nldm, **ldmarr;
    int		i, len;
    char	*type, *value;
    int		rc = 0;

    Debug( LDAP_DEBUG_ARGS, "Trace: (%d) ***** ldap_op: operation = modify\n",
	    getpid(), 0, 0 );

    state = AWAITING_OP;
    nvals = 0;
    nops = 0;
    ldmarr = NULL;

    /*
     * Construct a null-terminated array of LDAPMod structs.
     */
    p = rpl->mods;
    for ( p = rpl->mods, i = 0; p[ i ].type != NULL; i++ ) {
	type = p[ i ].type;
	value = p[ i ].value;
	len = p[ i ].len;
	Debug( LDAP_DEBUG_ARGS, "Trace: (%d) *** op_ldap_modify: %s: %s\n",
	    getpid(), type, value ? value : "" );
	switch ( getmodtype( type )) {
	case T_MODSEP:
	    state = T_MODSEP; /* Got a separator line "-\n" */
	    continue;
	case T_MODOPADD:
	    state = T_MODOPADD;
	    ldmarr = ( LDAPMod ** )
		    ch_realloc(ldmarr, (( nops + 2 ) * ( sizeof( LDAPMod * ))));
	    ldmarr[ nops ] = ldm = alloc_ldapmod();
	    ldm->mod_op = LDAP_MOD_ADD | LDAP_MOD_BVALUES;
	    ldm->mod_type = value;
	    nvals = 0;
	    nops++;
	    break;
	case T_MODOPREPLACE:
	    state = T_MODOPREPLACE;
	    ldmarr = ( LDAPMod ** )
		    ch_realloc(ldmarr, (( nops + 2 ) * ( sizeof( LDAPMod * ))));
	    ldmarr[ nops ] = ldm = alloc_ldapmod();
	    ldm->mod_op = LDAP_MOD_REPLACE | LDAP_MOD_BVALUES;
	    ldm->mod_type = value;
	    nvals = 0;
	    nops++;
	    break;
	case T_MODOPDELETE:
	    state = T_MODOPDELETE;
	    ldmarr = ( LDAPMod ** )
		    ch_realloc(ldmarr, (( nops + 2 ) * ( sizeof( LDAPMod * ))));
	    ldmarr[ nops ] = ldm = alloc_ldapmod();
	    ldm->mod_op = LDAP_MOD_DELETE | LDAP_MOD_BVALUES;
	    ldm->mod_type = value;
	    nvals = 0;
	    nops++;
	    break;
	default:
	    if ( state == AWAITING_OP ) {
		Debug( LDAP_DEBUG_ANY,
			"Error: (%d) op_ldap_modify: unknown mod type %s\n",
			getpid(), type, 0 );
		continue;
	    }

	    /*
	     * We should have an attribute: value pair here.
	     * Construct the mod_bvalues part of the ldapmod struct.
	     */
	    Debug( LDAP_DEBUG_ARGS,
		    "Trace: (%d) op_ldap_modify: berval for %s\n",
		    getpid(), type, 0 );
	    if ( strcasecmp( type, ldm->mod_type )) {
		Debug( LDAP_DEBUG_ANY,
			"Error: malformed modify op, %s: %s (expecting %s:)\n",
			type, value, ldm->mod_type );
		continue;
	    }
	    ldm->mod_bvalues = ( struct berval ** )
		    ch_realloc( ldm->mod_bvalues,
		    ( nvals + 2 ) * sizeof( struct berval * ));
	    ldm->mod_bvalues[ nvals + 1 ] = NULL;
	    ldm->mod_bvalues[ nvals ] = ( struct berval * )
		    ch_malloc( sizeof( struct berval ));
	    ldm->mod_bvalues[ nvals ]->bv_val = value;
	    ldm->mod_bvalues[ nvals ]->bv_len = len;
	    nvals++;
	}
    }
    ldmarr[ nops ] = NULL;

    if ( nops > 0 ) {
	/* Actually perform the LDAP operation */
	dump_ldm_array( ldmarr );

	rc = ldap_modify_s( replica->repl_ldp, rpl->dn, ldmarr );

	if ( rc != LDAP_SUCCESS ) {
	    Debug( LDAP_DEBUG_ANY, "Error: (%d) ldap_modify_s failed: %s\n",
		    getpid(), ldap_err2string( rc ), 0 );
	}
    }
    free_ldmarr( ldmarr );
    return( rc );
}




/*
 * Perform an ldap delete operation.
 */
static int
op_ldap_delete(
replicainfo *replica,
struct repl *rpl )
{
    int		rc;

    Debug( LDAP_DEBUG_ARGS, "Trace (%d): ***** ldap_delete_s( ldp, \"%s\")\n",
	    getpid(), rpl->dn, 0 );

    rc = ldap_delete_s( replica->repl_ldp, rpl->dn );

    if ( rc != LDAP_SUCCESS ) {
	Debug( LDAP_DEBUG_ANY, "Error: (%d) ldap_delete_s failed: %s\n",
		getpid(), ldap_err2string( rc ), 0 );
    }
    return( rc );
}




/*
 * Perform an ldap modrdn operation.
 */
static int
op_ldap_modrdn(
replicainfo *replica,
struct repl *rpl )
{
    int		rc = 0;
    modinfo	*p;

    p = rpl->mods;
    if ( p == NULL || strcmp( p->type, T_NEWRDNSTR )) {
	Debug( LDAP_DEBUG_ANY, "Error: (%d) op_ldap_modrdn: bad type \"%s\\n",
		getpid(), p->type, 0 );
	return;
    }
    if ( p->value == NULL || p->value[0] == '\0' ) {
	Debug( LDAP_DEBUG_ANY,
		"Error: op_ldap_modrdn: (%d) newrdn is empty\n",
		getpid(), 0, 0 );
	return;
    }
    Debug( LDAP_DEBUG_ARGS,
	    "Trace: (%d) ***** ldap_modrdn_s( ldp, \"%s\", \"%s\")\n",
	    getpid(), rpl->dn, p->value );

    rc = ldap_modrdn_s( replica->repl_ldp, rpl->dn, p->value );

    if ( rc != LDAP_SUCCESS ) {
	Debug( LDAP_DEBUG_ANY, "Error: (%d) ldap_modrdn_s failed: %s\n",
		getpid(), ldap_err2string( rc ), 0 );
    }
    return( rc );
}



/*
 * Allocate and initialize an ldapmod struct.
 */
static LDAPMod *
alloc_ldapmod()
{
    LDAPMod	*ldm;

    ldm = ( struct ldapmod * ) ch_malloc( sizeof ( struct ldapmod ));
    ldm->mod_type = NULL;
    ldm->mod_bvalues = ( struct berval ** ) NULL;
    return( ldm );
}



/*
 * Free an ldapmod struct associated mod_bvalues.  NOTE - it is assumed
 * that mod_bvalues and mod_type contain pointers to the same block of memory
 * pointed to by the repl struct.  Therefore, it's not freed here.
 */
static void
free_ldapmod(
LDAPMod *ldm )
{
    int		i;

    if ( ldm == NULL ) {
	return;
    }
    if ( ldm->mod_bvalues != NULL ) {
	for ( i = 0; ldm->mod_bvalues[ i ] != NULL; i++ ) {
	    free( ldm->mod_bvalues[ i ] );
	}
    }
    free( ldm );
    return;
}


/*
 * Free an an array of LDAPMod pointers and the LDAPMod structs they point
 * to.
 */
static void
free_ldmarr(
LDAPMod **ldmarr )
{
    int	i;

    for ( i = 0; ldmarr[ i ] != NULL; i++ ) {
	free_ldapmod( ldmarr[ i ] );
    }
    free( ldmarr );
}


/*
 * Create a berval with a single value. 
 */
static struct berval **
make_singlevalued_berval( 
char	*value,
int	len )
{
    struct berval **p;

    p = ( struct berval ** ) ch_malloc( 2 * sizeof( struct berval * ));
    p[ 0 ] = ( struct berval * ) ch_malloc( sizeof( struct berval ));
    p[ 1 ] = NULL;
    p[ 0 ]->bv_val = value;
    p[ 0 ]->bv_len = len;
    return( p );
}


/*
 * Given a modification type (string), return an enumerated type.
 * Avoids ugly copy in op_ldap_modify - lets us use a switch statement
 * there.
 */
static int
getmodtype( 
char *type )
{
    if ( !strcmp( type, T_MODSEPSTR )) {
	return( T_MODSEP );
    }
    if ( !strcmp( type, T_MODOPADDSTR )) {
	return( T_MODOPADD );
    }
    if ( !strcmp( type, T_MODOPREPLACESTR )) {
	return( T_MODOPREPLACE );
    }
    if ( !strcmp( type, T_MODOPDELETESTR )) {
	return( T_MODOPDELETE );
    }
    return( T_ERR );
}


/*
 * Perform an LDAP unbind operation.  If replica is NULL, or the
 * repl_ldp is NULL, just return LDAP_SUCCESS.  Otherwise, unbind,
 * set the ldp to NULL, and return the result of the unbind call.
 */
int
do_unbind(
replicainfo *replica )
{
    int		rc = LDAP_SUCCESS;

    if (( replica != NULL ) && ( replica->repl_ldp != NULL )) {
	rc = ldap_unbind( replica->repl_ldp );
	if ( rc != LDAP_SUCCESS ) {
	    Debug( LDAP_DEBUG_ANY,
		    "Error: (%d) do_bind: ldap_unbind failed: %s\n",
		    getpid(), ldap_err2string( rc ), 0 );
	}
	replica->repl_ldp = NULL;
    }
    return rc;
}



/*
 * Perform an LDAP bind operation to the replication site given
 * by replica.  If replica->repl_ldp is non-NULL, then we unbind
 * from the replica before rebinding.  It should be safe to call
 * this to re-connect if the replica's connection goes away
 * for some reason. XXX - kerberos bind *not finished*.
 */
int
do_bind( 
replicainfo *replica )
{
    int		rc;
    char	msgbuf[ 1024];
#ifdef KERBEROS
    int retval = 0;
    int kni, got_tgt;
    char **krbnames;
    char *skrbnames[ 2 ];
    char realm[ REALM_SZ ];
    char name[ ANAME_SZ ];
    char instance[ INST_SZ ];
#endif /* KERBEROS */


    if ( replica == NULL ) {
	Debug( LDAP_DEBUG_ANY, "Error: (%d) do_bind: null replica ptr\n",
		getpid(), 0, 0 );
	return( BIND_ERR_BADLDP );
    }

    sprintf( msgbuf, "bind to server %s:%d as \"%s\", pw = \"%s\"",
	    replica->repl_hostname, replica->repl_port,
	    replica->repl_bind_dn, replica->repl_password );
    Debug( LDAP_DEBUG_ARGS, "Trace: do_bind: (%d) %s\n", getpid(), msgbuf, 0 );

    if ( replica->repl_ldp != NULL ) {
	Debug( LDAP_DEBUG_ARGS, "Trace: do_bind: repl_ldp was non-null\n",
		getpid(), 0, 0 );
	rc = ldap_unbind( replica->repl_ldp );
	if ( rc != LDAP_SUCCESS ) {
	    Debug( LDAP_DEBUG_ANY,
		    "Error: (%d) do_bind: ldap_unbind failed: %s\n",
		    getpid(), ldap_err2string( rc ), 0 );
	}
	replica->repl_ldp = NULL;
    }

    replica->repl_ldp = ldap_open( replica->repl_hostname, replica->repl_port );
    if ( replica->repl_ldp == NULL ) {
	Debug( LDAP_DEBUG_ANY, "Error (%d): ldap_open() failed: %s\n",
		getpid(), sys_errlist[ errno ], 0 );
	return( BIND_ERR_OPEN );
    }

    switch ( replica->repl_bind_method ) {
    case AUTH_KERBEROS:
#ifndef KERBEROS
	Debug( LDAP_DEBUG_ANY,
	    "Error (%d): Kerberos bind used, but not compiled w/kerberos\n",
	    getpid(), 0, 0 );
	return( BIND_ERR_KERBEROS_FAILED );
#else /* KERBEROS */
	/*
	 * Bind using kerberos.
	 * If "bindprincipal" was given in the config file, then attempt
	 * to get a TGT for that principal (via the srvtab file).  If only
	 * a binddn was given, then we need to read that entry to get
	 * the kerberosName attributes, and try to get a TGT for one
	 * of them.  All are tried.  The first one which succeeds is
	 * returned.  XXX It might be a good idea to just require a
	 * bindprincipal.  Reading the entry every time might be a significant
	 * amount of overhead, if the connection is closed between most
	 * updates.
	 */

	if ( replica->repl_prin != NULL ) {
	    skrbnames[ 0 ] = replica->repl_prin;
	    skrbnames[ 1 ] = NULL;
	    krbnames = skrbnames;
	} else {
	    krbnames = read_krbnames( replica );
	}	
	    
	if (( krbnames == NULL ) || ( krbnames[ 0 ] == NULL )) {
	    Debug( LDAP_DEBUG_ANY,
		    "Error (%d): Can't find krbname for binddn \"%s\"\n",
		    getpid(), replica->repl_bind_dn, 0 );
	    retval = BIND_ERR_KERBEROS_FAILED;
	    goto kexit;
	}
	/*
	 * Now we've got one or more kerberos principals.  See if any
	 * of them are in the srvtab file.
	 */
	got_tgt = 0;
	for ( kni = 0; krbnames[ kni ] != NULL; kni++ ) {
	    rc = kname_parse( name, instance, realm, krbnames[ kni ]);
	    if ( rc != KSUCCESS ) {
		continue;
	    }
	    upcase( realm );
	    rc = krb_get_svc_in_tkt( name, instance, realm, "krbtgt", realm,
		    1, replica->repl_srvtab );
	    if ( rc != KSUCCESS) {
		Debug( LDAP_DEBUG_ANY, "Error (%d): Can't get TGT for %s: %s\n",
			getpid(), krbnames[ kni ], krb_err_txt[ rc ] );
	    } else {
		got_tgt = 1;
		break;
	    }
	}
	if (!got_tgt) {
	    Debug( LDAP_DEBUG_ANY,
		    "Error (%d): Could not obtain TGT for DN \"%s\"\n", 
		    getpid(), replica->repl_bind_dn, 0 );
	    retval = BIND_ERR_KERBEROS_FAILED;
	    goto kexit;
	}
	/*
	 * We've got a TGT.  Do a kerberos bind.
	 */
	rc = ldap_kerberos_bind1_s( replica->repl_ldp, replica->repl_bind_dn );
	replica->repl_prin = strdup( krbnames[ kni ] );
	if ( rc != LDAP_SUCCESS ) {
	    Debug( LDAP_DEBUG_ANY, "Error (%d): kerberos bind failed: %s\n",
		getpid(), ldap_err2string( rc ), 0 );
	    retval = BIND_ERR_KERBEROS_FAILED;
	    goto kexit;
	}
kexit:	if ( krbnames != NULL ) {
	    ldap_value_free( krbnames );
	}
	return( retval);
	break;
#endif /* KERBEROS */
    case AUTH_SIMPLE:
	/*
	 * Bind with a plaintext password.
	 */
	rc = ldap_simple_bind_s( replica->repl_ldp, replica->repl_bind_dn,
		replica->repl_password );
	if ( rc != LDAP_SUCCESS ) {
	    Debug( LDAP_DEBUG_ANY,
		    "Error (%d): ldap_simple_bind_s failed: %s\n",
		    getpid(), ldap_err2string( rc ), 0 );
	    return( BIND_ERR_SIMPLE_FAILED );
	} else {
	    return( BIND_OK );
	}
	break;
    default:
	Debug(  LDAP_DEBUG_ANY,
		"Error: (%d) do_bind: unknown auth type \"%d\"\n",
		getpid(), replica->repl_bind_method, 0 );
	return( BIND_ERR_BAD_ATYPE );
    }
}





/*
 * For debugging.  Print the contents of an ldmarr array.
 */
static void
dump_ldm_array(
LDAPMod **ldmarr )
{
    int			 i, j;
    LDAPMod		*ldm;
    struct berval	*b;
    char		*msgbuf;

    for ( i = 0; ldmarr[ i ] != NULL; i++ ) {
	ldm = ldmarr[ i ];
	Debug( LDAP_DEBUG_ARGS,
		"Trace (%d): *** ldmarr[ %d ] contents:\n",
		getpid(), i, 0 );
	Debug( LDAP_DEBUG_ARGS,
		"Trace (%d): *** ldm->mod_op: %d\n",
		getpid(), ldm->mod_op, 0 );
	Debug( LDAP_DEBUG_ARGS,
		"Trace (%d): *** ldm->mod_type: %s\n",
		getpid(), ldm->mod_type, 0 );
	if ( ldm->mod_bvalues != NULL ) {
	    for ( j = 0; ( b = ldm->mod_bvalues[ j ] ) != NULL; j++ ) {
		msgbuf = ch_malloc( b->bv_len + 512 );
		sprintf( msgbuf, "***** bv[ %d ] len = %d, val = <%s>",
			j, b->bv_len, b->bv_val );
		Debug( LDAP_DEBUG_ARGS,
			"Trace (%d):%s\n", getpid(), msgbuf, 0 );
		free( msgbuf );
	    }
	}
    }
}


/*
 * Get the kerberos names from the binddn for "replica" via an ldap search.
 * Returns a null-terminated array of char *, or NULL if the entry could
 * not be found or there were no kerberosName attributes.  The caller is
 * responsible for freeing the returned array and strings it points to.
 */
static char **read_krbnames(
replicainfo *replica )
{
    int rc;
    char **krbnames;
    int ne;
    LDAPMessage *result, *entry;

    /* First need to bind as NULL */
    rc = ldap_simple_bind_s( replica->repl_ldp, NULL, NULL );
    if ( rc != LDAP_SUCCESS ) {
	Debug( LDAP_DEBUG_ANY,
		"Error (%d): ldap_simple_bind_s (NULL) failed: %s\n",
		getpid(), sys_errlist[ errno ], 0 );
	return( NULL );
    }
    rc = ldap_search_st( replica->repl_ldp, replica->repl_bind_dn,
	    LDAP_SCOPE_BASE, "objectclass=*", kattrs, 0, &kst,
	    &result );
    if ( rc != LDAP_SUCCESS ) {
	Debug( LDAP_DEBUG_ANY,
		"Error (%d): ldap_search_st (NULL) failed: %s\n",
		getpid(), sys_errlist[ errno ], 0 );
	return( NULL );
    }
    ne = ldap_count_entries( replica->repl_ldp, result );
    if ( ne == 0 ) {
	Debug( LDAP_DEBUG_ANY,
		"Error (%d): Can't find entry for kerberos binddn \"%s\"\n",
		getpid(), replica->repl_bind_dn, 0 );
	    return( NULL );
    }
    if ( ne > 1 ) {
	Debug( LDAP_DEBUG_ANY,
		"Error (%d): Kerberos binddn \"%s\" is ambiguous\n",
		getpid(), replica->repl_bind_dn, 0 );
	    return( NULL );
    }
    entry = ldap_first_entry( replica->repl_ldp, result );
    if ( entry == NULL ) {
	Debug( LDAP_DEBUG_ANY,
		"Error (%d): Can't find entry for kerberos binddn \"%s\"\n",
		    getpid(), replica->repl_bind_dn, 0 );
	return( NULL );
    }
    krbnames = ldap_get_values( replica->repl_ldp, entry, "kerberosName" );
    ldap_msgfree( result );
    return( krbnames );
}



/*
 * upcase a string
 */
void
upcase(
char *s )
{
    char *p;

    for ( p = s; ( p != NULL ) && ( *p != '\0' ); p++ ) {
	if ( islower( *p )) {
	    *p = toupper( *p );
	}
    }
}
