/*
 * Copyright (c) 1990, 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */


#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include "portable.h"
#include "../slapd/slap.h"
#include "slurp.h"
#include <ldapconfig.h>

/*
 * Globals
 */
/* Debugging level for LDAP library and slurpd */
int	ldap_debug = 0;
/* The name of the slapd config file (which is also our config file */
char	*slapd_configfile = SLAPD_DEFAULT_CONFIGFILE;
/* if ldap_syslog != 0,  Debug() output goes to syslog, otherwise to stderr */
int	ldap_syslog = 0;
/* The LDAP syslog level, e.g. LOG_DEBUG, etc. */
int	ldap_syslog_level = LOG_DEBUG;
/* How long the master slurpd sleeps when there's no work to do */
int	no_work_interval = DEFAULT_NO_WORK_INTERVAL;
/* Linked list of info about replicas we're servicing */
struct replicainfo
	*replicas = NULL;
/* We keep running until slurpd_shutdown is nonzero.  A HUP signal set this */
int	slurpd_shutdown = 0;
/* Directory where our replica files are written/read */
char	*slurpd_rdir = DEFAULT_SLURPD_REPLICA_DIR;
/* Name of the replication log slapd is writing (and we are reading) */
char	slapd_replogfile[ MAXPATHLEN ] = "\0";
/* Non-zero if we were given a replog file to process on command-line */
int	one_shot_mode = 0;

#ifdef KERBEROS
/* Default name of kerberos srvtab file */
char	*global_srvtab = SRVTAB;
#endif /* KERBEROS */

/*
 * Extern functions
 */
extern int copy_replog( char *, char * );
extern void service_replicas( char *, char * );
extern void reap_child();

/*
 * Forward declarations
 */
static int usage( char * );





main( int argc, char ** argv )
{
    int		i;
    char	*myname;
    extern char	*optarg;
    char	priv_replogfile[ MAXPATHLEN ]; /* our copy of replog file */

    while ( (i = getopt( argc, argv, "hd:f:r:t:k:o" )) != EOF ) {
	switch ( i ) {
#ifdef LDAP_DEBUG
	case 'd':	/* turn on debugging */
	    if ( optarg[0] == '?' ) {
		printf( "Debug levels:\n" );
		printf( "\tLDAP_DEBUG_TRACE\t%d\n",
			LDAP_DEBUG_TRACE );
		printf( "\tLDAP_DEBUG_PACKETS\t%d\n",
			LDAP_DEBUG_PACKETS );
		printf( "\tLDAP_DEBUG_ARGS\t\t%d\n",
			LDAP_DEBUG_ARGS );
		printf( "\tLDAP_DEBUG_CONNS\t%d\n",
			LDAP_DEBUG_CONNS );
		printf( "\tLDAP_DEBUG_BER\t\t%d\n",
			LDAP_DEBUG_BER );
		printf( "\tLDAP_DEBUG_FILTER\t%d\n",
			LDAP_DEBUG_FILTER );
		printf( "\tLDAP_DEBUG_CONFIG\t%d\n",
			LDAP_DEBUG_CONFIG );
		printf( "\tLDAP_DEBUG_ACL\t\t%d\n",
			LDAP_DEBUG_ACL );
		printf( "\tLDAP_DEBUG_ANY\t\t%d\n",
			LDAP_DEBUG_ANY );
		exit( 0 );
	    } else {
		ldap_debug = atoi( optarg );
	    }
	    break;
#else /* LDAP_DEBUG */
	case 'd':	/* can't enable debugging - not built with debug code */
	    fprintf( stderr, "must compile with LDAP_DEBUG for debugging\n" );
	    break;
#endif /* LDAP_DEBUG */
	case 'f':	/* slapd config file */
	    slapd_configfile = strdup( optarg );
	    break;
	case 'r':	/* slapd replog file */
	    strcpy( slapd_replogfile, optarg );
	    break;
	case 't':	/* dir to use for our copies of replogs */
	    slurpd_rdir = strdup( optarg );
	    break;
	case 'k':	/* name of kerberos srvtab file */
#ifdef KERBEROS
	    global_srvtab = strdup( optarg );
#else /* KERBEROS */
	    fprintf( stderr, "must compile with KERBEROS to use -k option\n" );
#endif /* KERBEROS */
	    break;
	case 'h':
	    usage( argv[0] );
	    exit( 1 );
	case 'o':
	    one_shot_mode = 1;
	    break;
	default:
	    usage( argv[0] );
	    exit( 1 );
	}
    }

    if ( (myname = strrchr( argv[0], '/' )) == NULL ) {
	myname = strdup( argv[0] );
    } else {
	myname = strdup( myname + 1 );
    }

    /* Set location/name of our private copy of the slapd replog file */
    sprintf( priv_replogfile, "%s/%s", slurpd_rdir, DEFAULT_SLURPD_REPLOGFILE );

    slurpd_read_config( slapd_configfile );

#ifdef LOG_LOCAL4
    openlog( myname, OPENLOG_OPTIONS, LOG_LOCAL4 );
#else
    openlog( myname, OPENLOG_OPTIONS );
#endif

    detach();

    (void) signal( SIGCHLD, reap_child );
    (void) signal( SIGUSR1, SIG_IGN ); /* so children inherit */

    service_replicas( slapd_replogfile, priv_replogfile );
}





static int
usage( char *name )
{
    fprintf( stderr, "usage: %s\t[-d debug-level] [-s syslog-level]\n", name );
    fprintf( stderr, "\t\t[-c slapd-config-file] [-r slapd-replog-file]\n" );
    fprintf( stderr, "\t\t[-t tmp-dir] [-o]\n" );
}

