/*
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <sys/stat.h>

#include "portable.h"
#include "slurp.h"

/* External references */
extern int	file_nonempty( char * );
extern struct repl *get_replog_entry_t( FILE *, struct timeval *, int * );
extern struct repl *get_replog_entry( FILE *, int * );
extern FILE *lock_fopen( char *, char *, FILE ** );
#ifndef SYSERRLIST_IN_STDIO
extern char *sys_errlist[];
#endif
extern int perform_ldap_operation( struct replicainfo *, struct repl * );
extern int slurpd_shutdown;
extern int relinquish_lock(char *, FILE *, FILE * );
extern int acquire_lock(char *, FILE **, FILE ** );
extern int do_bind( replicainfo * );
extern int do_unbind( replicainfo * );
extern char *slurpd_rdir;


/* Local variables */
static int	work_waiting;
static int	got_usr1;
static int	cf_warning;	/* non-zero if we've already logged a warning */
				/* about this replica being unavailable       */

/* Forward references */
static struct timeval *get_lastoptime( FILE * );
struct timeval *str2optime( char * );
static void update_lastfile( char *, struct repl * );
static void warn_cf( int );
static void usr1_handler();
static void write_reject( replicainfo *, struct repl *, char * );
static int write_repl( struct repl *, FILE * );
static struct repl *get_next_repl( char *, off_t * );


int replicate(
    replicainfo *replica )
{
    char	replicafile[ MAXPATHLEN ];
    char	lastfile[ MAXPATHLEN ];
    off_t	offset;
    struct repl	*rpl;
    int		rc;

    /* Construct filenames */
    sprintf( replicafile, "%s/%s:%d.replog", DEFAULT_SLURPD_REPLICA_DIR,
	    replica->repl_hostname, replica->repl_port );
    strcpy( lastfile, replicafile );
    strcat( lastfile, ".last" );

    offset = (off_t) 0L;

    for (;;) {
	rpl = get_next_repl( replicafile, &offset );
	if ( rpl == NULL ) {
	    got_usr1 = 0;
	    signal( SIGUSR1, usr1_handler );
	    sleep( REPLICA_SLEEP_TIME );
	    signal( SIGUSR1, SIG_IGN );
	    if ( !got_usr1 ) {
		return( 0 );
	    }
	} else {
	    rc = perform_ldap_operation( replica, rpl );
	    if ( rc == LDAP_SERVER_DOWN ) {
		/*
		 * Try to re-establish connection - perform_ldap_operation()
		 * will rebind if repl_ldp is NULL.
		 */
		replica->repl_ldp = NULL;
		rc = perform_ldap_operation( replica, rpl );
	    }
	    if ( rc != LDAP_SUCCESS ) {
		write_reject( replica, rpl, ldap_err2string( rc ));
	    } else {
		update_lastfile( lastfile, rpl );
	    }
	}
    }
}



/*
 * Get the next replication record at offset bytes in file replicafile.
 * If offset is zero, check for the existence of a "{replicafile}.last"
 * file, and if it exists, assume that it contains the timestamp of the
 * last successful replication record.  The offset of the *next* replication
 * record past the one returned is stored in the variable pointed to
 * by offsetp.
 */
static struct repl *
get_next_repl(
char	*replicafile,
off_t	*offsetp )
{
    FILE *rfp, *lfp, *lafp, *lalfp;
    struct timeval *lastoptime;
    struct repl *rpl;
    char lastfile[ MAXPATHLEN ];
    int status;

    /* Construct lastfile name */
    strcpy( lastfile, replicafile );
    strcat( lastfile, ".last" );

    if ( *offsetp == (off_t) 0L ) {
	/* See if there's a lastfile around */
	if (( lafp = fopen( lastfile, "r" )) == NULL ) {
	    /* No lastfile.  Create it */
	    int fd;
	    Debug( LDAP_DEBUG_ARGS, "Trace: replicate (%d): creating %s\n",
		    getpid(), lastfile, 0 );
	    fd = open( lastfile, O_WRONLY | O_CREAT | O_TRUNC, 0660 );
	    if ( fd < 0 ) {
		Debug( LDAP_DEBUG_ANY,
			"Error: replicate (%d): error creating %s: %s\n",
			getpid(), lastfile, sys_errlist[ errno ] );
	    } else {
		close( fd );
	    }
	}    
	/*
	 * Retrieve the first replication record whose timestamp is
	 * greater than the time given in the lastfile.
	 */
	if ( acquire_lock( lastfile, &lafp, &lalfp ) < 0 ) {
	    return( NULL );
	}
	lastoptime = get_lastoptime( lafp );
	if ( acquire_lock( replicafile, &rfp, &lfp ) < 0 ) {
	    Debug( LDAP_DEBUG_ARGS,
		    "Trace: replicate (%d): unable to lock \"%s\"\n",
		    getpid(), replicafile, 0 );
	    return( NULL );
	}
	rpl = get_replog_entry_t( rfp, lastoptime, &status );
	*offsetp = ftell( rfp );
	(void) relinquish_lock( replicafile, rfp, lfp );
	(void) relinquish_lock( lastfile, lafp, lalfp );
	return( rpl );
    }

    /*
     * We were given an offset, so open the file, seek to that offset,
     * read the replication record, and return the record and update the
     * offset.
     */
    if ( acquire_lock( replicafile, &rfp, &lfp ) < 0 ) {
	Debug( LDAP_DEBUG_ARGS,
		"Trace: replicate (%d): unable to lock \"%s\"\n",
		getpid(), replicafile, 0 );
	return( NULL );
    }
    fseek( rfp, *offsetp, 0 );
    rpl = get_replog_entry( rfp, &status );
    *offsetp = ftell( rfp );
    (void) relinquish_lock( replicafile, rfp, lfp );
    return( rpl );
}
    

	
static void
usr1_handler()
{
    got_usr1 = 1;
}



/*
 * Given an open file descriptor, get the first line in the file
 * and convert it to a timeval.  Returns a pointer to a static area.
 *
 * Lines in the last file should be of the format S.U
 * where "S" is a "unix time" value and "U" is a four-digit
 * "disambiguator", left-padded with zeros.  This "disambiguator
 * value is used for the tv_usec member of the timeval struct.
 */
static struct timeval
*get_lastoptime(
FILE	*lafp )
{
    static struct timeval tv;
    char buf[ 80 ];

    Debug( LDAP_DEBUG_ARGS,
	    "Trace: get_lastoptime (%d): begin\n", getpid(), 0, 0 );

    if ( lafp == NULL ) {
	tv.tv_sec = tv.tv_usec = 0L;
	Debug( LDAP_DEBUG_ANY, "Error: get_lastoptime (%d): null lafp\n",
		getpid(), 0, 0 );
	return( &tv );
    }

    fseek( lafp, 0L, 0 );
    if ( fgets( buf, sizeof( buf ), lafp ) == NULL ) {
	tv.tv_sec = tv.tv_usec = 0L;
	Debug( LDAP_DEBUG_ARGS, "Trace: get_lastoptime (%d): null read\n",
		getpid(), 0, 0 );
	return( &tv );
    }
    Debug( LDAP_DEBUG_ARGS, "Trace: get_lastoptime (%d): returning %ld.%ld\n",
	    getpid(), tv.tv_sec, tv.tv_usec );
    return( str2optime( buf ));
}



struct timeval
*str2optime( 
char *s )
{
    static struct timeval tv;
    char *p;

    if (( s == NULL ) || ( *s == '\0' )) {
	tv.tv_sec = tv.tv_usec = 0L;
	return( &tv );
    }
    p = strchr( s, '.' );
    if ( p != NULL ) {
	*p++ = '\0';
	tv.tv_sec = atol( s );
	tv.tv_usec = atol( p );
    } else {
	tv.tv_sec = atol( s );
	tv.tv_usec = 0L;
    }
    return( &tv );
}



static void
trim_replicafile(
char		*replicafile,
struct timeval	*lastoptime )
{
    Debug( LDAP_DEBUG_ARGS, "Trace (%d): trim_replicafile: begin\n",
	    getpid(), 0, 0);
}


/*
 * Write the "time" field of repl to the given fp
 */
static void
update_lastfile(
char	*lastfile,
struct repl *rpl )
{
    FILE *lafp, *lalfp;

    if ( rpl == NULL ) {
	return;
    }

Debug( LDAP_DEBUG_ANY, "update_lastfile(%d): about to lock \"%s\"\n",
	getpid(), lastfile, 0 );
    if ( acquire_lock( lastfile, &lafp, &lalfp ) < 0 ) {
	return;
    }
Debug( LDAP_DEBUG_ANY, "update_lastfile(%d): got lock on \"%s\"\n",
	getpid(), lastfile, 0 );

    fputs( rpl->time, lafp );
    fflush( lafp );
Debug( LDAP_DEBUG_ANY, "update_lastfile(%d): about to unlock \"%s\"\n",
	getpid(), lastfile, 0 );
    (void) relinquish_lock( lastfile, lafp, lalfp );
Debug( LDAP_DEBUG_ANY, "update_lastfile(%d): unlocked \"%s\"\n",
	getpid(), lastfile, 0 );
}


/*
 * Given an enumerated error, generate a warning message
 */
static void
warn_cf( 
int code )
{
    switch ( code ) {
    case BIND_OK:
	Debug( LDAP_DEBUG_ARGS, "Trace (%d) bind succeeded\n",
		getpid(), 0, 0 );
	break;
    case BIND_ERR_BADLDP: 
	Debug( LDAP_DEBUG_ANY, "Error (%d): bind failed - null LDAP pointer\n",
		getpid(), 0, 0);
	break;
    case BIND_ERR_OPEN:
	Debug( LDAP_DEBUG_ANY, "Error (%d): ldap_open() failed\n",
		getpid(), 0, 0 );
	break;
    case BIND_ERR_BAD_ATYPE:
	Debug( LDAP_DEBUG_ANY,
		"Error (%d): bind failed - unknown authentication type\n",
		getpid(), 0, 0 );
	break;
    case BIND_ERR_SIMPLE_FAILED:
	Debug( LDAP_DEBUG_ANY, "Error (%d): ldap_simple_bind() failed\n",
		getpid(), 0, 0 );
	break;
    default:
	Debug( LDAP_DEBUG_ANY, "Error (%d): do_bind failed, error = %d\n",
		getpid(), code, 0 );
    }

}


/*
 * Write a replication record to a reject file.  The reject file has the
 * same name as the replica's private copy of the file but with ".rej"
 * appended (e.g. "/usr/tmp/<hostname>:<port>.rej")
 */
static void
write_reject(
replicainfo *replica,
struct repl *rpl,
char *msg )
{
    char	rejfile[ MAXPATHLEN ];
    FILE	*rfp, *lfp;
    int		rc;

    sprintf( rejfile, "%s/%s:%d.rej", slurpd_rdir, replica->repl_hostname,
	    replica->repl_port );

    if ( access( rejfile, F_OK ) < 0 ) {
	/* Doesn't exist - try to create */
	int rjfd;
	if (( rjfd = open( rejfile, O_RDWR | O_APPEND | O_CREAT,
		S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP )) < 0 ) {
	    Debug( LDAP_DEBUG_ANY,
		"Error: write_reject (%d): Cannot create \"%s\": %s\n",
		getpid(), rejfile, sys_errlist[ errno ] );
	    return;
	} else {
	    close( rjfd );
	}
    }
    if (( rc = acquire_lock( rejfile, &rfp, &lfp )) < 0 ) {
	Debug( LDAP_DEBUG_ANY, "Error (%d): cannot open reject file \"%s\"\n",
		getpid(), rejfile, 0 );
    } else {
	fseek( rfp, 0, 2 );
	fprintf( rfp, "%s: %s\n", ERROR_STR, msg );
	if ((rc = write_repl( rpl, rfp )) < 0 ) {
	    Debug( LDAP_DEBUG_ANY,
		    "Error (%d): cannot write reject file \"%s\"\n",
		    getpid(), rejfile, 0 );
	}
	(void) relinquish_lock( rejfile, rfp, lfp );
	Debug( LDAP_DEBUG_ANY,
		"Error (%d): ldap operation failed, data written to \"%s\"\n",
		getpid(), rejfile, 0 );
    }

}



/* 
 * Given a struct repl and a file pointer, write a replication record to
 * the file pointer.
 */
static int
write_repl(
struct repl *rpl,
FILE *fp )
{
    int		i;
    char	*s;

    for ( i = 0; rpl->replicas[ i ] != NULL; i++ ) {
	fprintf( fp, "replica: %s\n", rpl->replicas[ i ] );
    }
    fprintf( fp, "time: %s\n", rpl->time );
    fprintf( fp, "dn: %s\n", rpl->dn );
    fprintf( fp, "changetype: " );
    switch ( rpl->changetype ) {
    case T_ADDCT:
	s = T_ADDCTSTR;
	break;
    case T_MODIFYCT:
	s = T_MODIFYCTSTR;
	break;
    case T_DELETECT:
	s = T_DELETECTSTR;
	break;
    case T_MODRDNCT:
	s = T_MODRDNCTSTR;
	break;
    default:
	s = "IllegalModifyType!!!";
    }
    fprintf( fp, "%s\n", s );
    for ( i = 0; (( rpl->mods != NULL )  && ( rpl->mods[ i ].type != NULL ));
	    i++ ) {
	if ( !strcmp( rpl->mods[ i ].type, T_MODSEPSTR )) {
	    fprintf( fp, "%s\n", T_MODSEPSTR );
	} else {
	    fprintf( fp, "%s: %s\n",
		    rpl->mods[ i ].type ? rpl->mods[ i ].type : "",
		    rpl->mods[ i ].value ? rpl->mods[ i ].value : "" );
	}
    }
    fprintf( fp, "\n" );
    fflush( fp );
}


